import static BCDKit.fromBCD;
import static BCDKit.toBCD;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

/************************************************************************
* The register set of a Maxim Integrated DS1307 Real-Time Clock (RTC) chip.
* This class is independent of any I2C code.
************************************************************************/
public class DS1307Registers {

    private static final int MAX_SECONDS = 59;
    private static final int MAX_MINUTES = 59;
    
    private static final int MIN_DAY_NUMBER = 1;
    private static final int MAX_DAY_NUMBER = 7;
    
    private static final int MIN_DATE = 1;
    private static final int MAX_DATE = 31;
    
    private static final int MIN_MONTH = 1;
    private static final int MAX_MONTH = 12;
    
    private static final int MAX_YEAR = 99;
    private final static byte AMPM_MASK = 0x40;
    
    private byte bcdSeconds; // 00.59
    private byte bcdMinutes; // 00..59
    private byte bcdHours; // 00..23 (01..12 in AM/PM mode)
    private byte dayNumber; // 01.07
    private byte bcdDate; // 01..31
    private byte bcdMonth; // 01..12
    private byte bcdYear; // 00..99
    private byte controlByte;
    
    /**
    * Default constructor. All registers set to defaults.
    */
    public DS1307Registers() {
    
        dayNumber = bcdDate = bcdMonth = 1;
    }
    
    /**
    * Copy the first 8 bytes of the argument byte array into the registers.
    *
    * @param bytes
    */
    public void setRegisters(final byte[] bytes) {
    
        if (bytes.length < 8) {
            throw new IllegalArgumentException("Byte array length too short (needs to be at least 8): " + bytes.length);
        }
        
        int i = 0;
        
        bcdSeconds = bytes[i++];
        validate("Seconds", "have", getSeconds(), 0, MAX_SECONDS);
        
        bcdMinutes = bytes[i++];
        validate("Minutes", "have", getMinutes(), 0, MAX_MINUTES);
        
        bcdHours = bytes[i++];
        
        dayNumber = bytes[i++];
        validate("Day number", "has", getDayNumber(), MIN_DAY_NUMBER, MAX_DAY_NUMBER);
        
        bcdDate = bytes[i++];
        validate("Date", "has", getDate(), MIN_DATE, MAX_DATE);
        
        bcdMonth = bytes[i++];
        validate("Month", "has", getMonth(), MIN_MONTH, MAX_MONTH);
        
        bcdYear = bytes[i++];
        validate("Year", "has", getYear(), 0, MAX_YEAR);
        
        controlByte = bytes[i++];
    }

    private void validate(final String name, final String hasHave, final int value, final int min, final int max) {
    
        if ((value < min) || (value > max)) {
            throw new IllegalArgumentException(name + " " + hasHave + " to be in [" + min + ".." + max + "] range. Illegal: " + value);
        }
    }
    
    /**
    * Get the register set as a raw byte array.
    *
    * @return a byte[8]
    */
    public byte[] asByteArray() {

        final byte[] bytes = new byte[8];
        bytes[0] = bcdSeconds;
        bytes[1] = bcdMinutes;
        bytes[2] = bcdHours;
        bytes[3] = dayNumber;
        bytes[4] = bcdDate;
        bytes[5] = bcdMonth;
        bytes[6] = bcdYear;
        bytes[7] = controlByte;

        return bytes;
    }
    
    public int getSeconds() {
        return fromBCD(bcdSeconds);
    }
    
    public void setSeconds(final int seconds) {
        bcdSeconds = toBCD(seconds);
    }
    
    public int getMinutes() {
        return fromBCD(bcdMinutes);
    }
    
    public void setMinutes(final int minutes) {
        bcdMinutes = toBCD(minutes);
    }
    
    public int getHours() {
        return fromBCD(bcdHours);
    }
    
    public void setHours(final int hours) {
        bcdHours = toBCD(hours);
    }
    
    public boolean isAMPMMode() {
        return (bcdHours & AMPM_MASK) != 0;
    }
    
    /**
    * Get the day number (1..7, representing Mon..Sun)
    *
    * @return the day number
    */
    public int getDayNumber() {
        return dayNumber;
    }
    
    public void setDayNumber(final int dayNumber) {
        this.dayNumber = (byte) dayNumber;
    }
    
    public int getDate() {
        return fromBCD(bcdDate);
    }
    
    public void setDate(final int date) {
        bcdDate = toBCD(date);
    }
    
    public int getMonth() {
        return fromBCD(bcdMonth);
    }
    
    public void setMonth(final int month) {
        bcdMonth = toBCD(month);
    }
    
    /**
    * Get the 2-digit year. I.e. 2014 is returned as 14.
    *
    * @return the year
    */
    public int getYear() {
        return fromBCD(bcdYear);
    }
    
    public void setYear(final int year) {
        bcdYear = toBCD(year);
    }
    
    public byte getControlByte() {
        return controlByte;
    }
    
    /**
    * Set the registers to reflect the given Date.
    *
    * @param date
    */
    @SuppressWarnings("deprecation")
    public void set(final Date date) {
        final int seconds = date.getSeconds();
        final int minutes = date.getMinutes();
        final int hours = date.getHours();
        
        final int day = date.getDate();
        final int month = date.getMonth();
        final int year = date.getYear();
        
        setSeconds(seconds);
        setMinutes(minutes);
        setHours(hours);
        
        setDate(day);
        setMonth(month + 1);
        setYear(year - 100);
    }
    
    public String asPrettyString() {
        final DateFormat format = new SimpleDateFormat("dd/MM/yyyy HH:mm:ss z");
        return formattedBy(format);
    }
    
    public String formattedBy(final DateFormat dateFormat) {
        final Date date = new D ate(100 + getYear(), getMonth() - 1, getDate(), getHours(), getMinutes(), getSeconds());
        return dateFormat.format(date);
    }
    
    @Override
    public String toString() {
        return "DS1307[" + //
                        hex(bcdSeconds) + hex(bcdMinutes) + hex(bcdHours) + "-" + //
                        hex(dayNumber) + "-" + //
                        hex(bcdDate) + hex(bcdMonth) + hex(bcdYear) + "-" + //
                        hex(controlByte) + //
                        "]";
    }
    
    private static String hex(final int byteValue) {
        return Integer.toHexString(byteValue | 0x80000000).substring(6, 8).toUpperCase();
    }
    
    public static void main(final String[] args) {
    
        final DS1307Registers ds1307Registers = new DS1307Registers();
        final byte[][] registerSetSamples = {
            new byte[] {
                0, 0, 0, 1, 1, 1, 0, 0
            }, new byte[] {
                0x00, 0x30, 0x14, 1, 0x02, 0x11, 0x14, 0
            }, new byte[] {
                0x59, 0x59, 0x23, 1, 0x31, 0x12, 0x79, 0
            }
        };
    
        for (final byte[] registerSetSample : registerSetSamples) {
            ds1307Registers.setRegisters(registerSetSample);
            System.out.println(ds1307Registers + " = " + ds1307Registers.asPrettyString());
        }
    }
}